defmodule ChronoscopeWeb.API.V1.NTS.KeyEstablishmentController do
  use ChronoscopeWeb, :controller

  require Logger

  alias Chronoscope.NTS

  @default_port 4460
  @max_host_length 255

  @nts Application.compile_env(:chronoscope, :nts, NTS)

  def get(conn, %{"host" => host, "port" => port}) do
    try do
      handle_get(conn, %{host: String.trim(host), port: String.to_integer(port)})
    rescue
      ArgumentError -> bad_request_response(conn, "invalid port")
    end
  end

  def get(conn, %{"host" => host}) do
    handle_get(conn, %{host: String.trim(host), port: @default_port})
  end

  def get(conn, _params) do
    bad_request_response(conn, "missing host")
  end

  defp handle_get(conn, %{host: ""}) do
    bad_request_response(conn, "empty host")
  end

  defp handle_get(conn, %{host: host, port: port}) when port > 0 and port < 65536 do
    case key_establishment_response(host, port) do
      {:ok, response} ->
        json(conn, %{status: :ok, response: format_response(response)})

      {:error, error} ->
        json(conn, %{status: :error, reason: to_string(error)})
    end
  end

  defp handle_get(conn, _params) do
    bad_request_response(conn, "port out of range")
  end

  defp key_establishment_response(host, port) do
    host
    |> String.slice(0, @max_host_length)
    |> @nts.key_establishment(port)
  end

  defp format_response(response) do
    response
    |> Map.take([:aead_algorithms, :cert_expiration, :cookie_length, :cookies, :next_protocols, :port, :server])
    |> Map.update(:cookies, 0, &length/1)
  end

  defp bad_request_response(conn, message) do
    conn
    |> put_status(:bad_request)
    |> json(%{error: message})
  end
end
