defmodule Chronoscope.NTSTest do
  use Chronoscope.Case, async: true

  alias Chronoscope.DynamicSupervisorMock
  alias Chronoscope.GenServerMock
  alias Chronoscope.NTS
  alias Chronoscope.RegistryMock

  import Mox

  setup :verify_on_exit!

  describe "Chronoscope.NTS.healthy?()" do
    test "is healthy" do
      assert NTS.healthy?() == true
    end
  end

  describe "Chronoscope.NTS.list()" do
    test "shows empty client list" do
      DynamicSupervisorMock
      |> expect(:which_children, fn _ -> [] end)

      assert NTS.list() == []
    end

    test "lists all children" do
      DynamicSupervisorMock
      |> expect(:which_children, fn _ -> [{1, 2, 3, 4}, {5, 6, 7, 8}] end)

      GenServerMock
      |> expect(:call, fn 2, :list -> :one end)
      |> expect(:call, fn 6, :list -> :two end)

      assert NTS.list() == [:one, :two]
    end
  end

  describe "Chronoscope.NTS.list_clients()" do
    test "shows empty client list" do
      assert NTS.list_clients([]) == []
    end

    test "lists the given clients and starts them if they don't exist" do
      RegistryMock
      |> expect(:lookup, fn _, "test:1" -> [{1, nil}] end)
      |> expect(:lookup, fn _, "test2:2" -> [] end)

      DynamicSupervisorMock
      |> expect(
        :start_child,
        fn Chronoscope.NTS.DynamicSupervisor,
           {Chronoscope.NTS.Client,
            [
              server: %{host: "test2", port: 2},
              name: {:via, RegistryMock, {Chronoscope.NTS.Registry, "test2:2"}}
            ]} ->
          {:ok, 2}
        end
      )

      GenServerMock
      |> expect(:call, fn 1, :list, 10_000 -> :one end)
      |> expect(:call, fn 2, :list, 10_000 -> :two end)

      assert NTS.list_clients([%{host: "test", port: 1}, %{host: "test2", port: 2}]) == [:one, :two]
    end
  end

  describe "Chronoscope.NTS.start_client()" do
    test "starts a client" do
      RegistryMock
      |> expect(:lookup, fn _, "test:1" -> [] end)

      DynamicSupervisorMock
      |> expect(
        :start_child,
        fn Chronoscope.NTS.DynamicSupervisor,
           {Chronoscope.NTS.Client,
            [
              server: %{host: "test", port: 1},
              name: {:via, RegistryMock, {Chronoscope.NTS.Registry, "test:1"}}
            ]} ->
          {:ok, 2}
        end
      )

      assert NTS.start_client(%{host: "test", port: 1}) == 2
    end

    test "returns an existing client" do
      RegistryMock
      |> expect(:lookup, fn _, "test:1" -> [{1, nil}] end)

      assert NTS.start_client(%{host: "test", port: 1}) == 1
    end
  end

  describe "Chronoscope.NTS.auto_refresh()" do
    test "sets up auto refresh for a new client" do
      RegistryMock
      |> expect(:lookup, fn _, "test:1" -> [] end)

      DynamicSupervisorMock
      |> expect(
        :start_child,
        fn Chronoscope.NTS.DynamicSupervisor,
           {Chronoscope.NTS.Client,
            [
              server: %{host: "test", port: 1},
              name: {:via, RegistryMock, {Chronoscope.NTS.Registry, "test:1"}}
            ]} ->
          {:ok, 2}
        end
      )

      GenServerMock
      |> expect(:call, fn 2, :auto_refresh, 10_000 -> :ok end)

      assert NTS.auto_refresh(%{host: "test", port: 1}) == :ok
    end

    test "sets up auto refresh for an existing client" do
      RegistryMock
      |> expect(:lookup, fn _, "test:1" -> [{2, nil}] end)

      GenServerMock
      |> expect(:call, fn 2, :auto_refresh, 10_000 -> :ok end)

      assert NTS.auto_refresh(%{host: "test", port: 1}) == :ok
    end
  end

  describe "Chronoscope.NTS.cancel_auto_refresh()" do
    test "cancels auto refresh for a new client" do
      RegistryMock
      |> expect(:lookup, fn _, "test:1" -> [] end)

      DynamicSupervisorMock
      |> expect(
        :start_child,
        fn Chronoscope.NTS.DynamicSupervisor,
           {Chronoscope.NTS.Client,
            [
              server: %{host: "test", port: 1},
              name: {:via, RegistryMock, {Chronoscope.NTS.Registry, "test:1"}}
            ]} ->
          {:ok, 2}
        end
      )

      GenServerMock
      |> expect(:call, fn 2, :cancel_auto_refresh, 10_000 -> :ok end)

      assert NTS.cancel_auto_refresh(%{host: "test", port: 1}) == :ok
    end

    test "cancels auto refresh for an existing client" do
      RegistryMock
      |> expect(:lookup, fn _, "test:1" -> [{2, nil}] end)

      GenServerMock
      |> expect(:call, fn 2, :cancel_auto_refresh, 10_000 -> :ok end)

      assert NTS.cancel_auto_refresh(%{host: "test", port: 1}) == :ok
    end
  end

  describe "Chronoscope.NTS.remove()" do
    test "does nothing if the client doesn't exist" do
      RegistryMock
      |> expect(:lookup, fn _, _ -> [] end)

      assert NTS.remove("localhost", 1111) == {:error, :not_found}
    end

    test "removes a client" do
      RegistryMock
      |> expect(:lookup, fn _, _ -> [{1, 2}] end)

      GenServerMock
      |> expect(:call, fn 1, :terminate, 10_000 -> :terminating end)

      assert NTS.remove("localhost", 1111) == {:ok, :terminating}
    end
  end

  describe "Chronoscope.NTS.key_establishment()" do
    test "creates a new client" do
      RegistryMock
      |> expect(:lookup, fn _, _ -> [] end)

      DynamicSupervisorMock
      |> expect(
        :start_child,
        fn Chronoscope.NTS.DynamicSupervisor,
           {Chronoscope.NTS.Client,
            [
              server: %{host: "localhost", port: 1111},
              name: {:via, RegistryMock, {Chronoscope.NTS.Registry, "localhost:1111"}}
            ]} ->
          {:ok, 1}
        end
      )

      GenServerMock
      |> expect(:call, fn 1, :key_establishment, 10_000 -> :result end)

      assert NTS.key_establishment("localhost", 1111) == :result
    end

    test "reuses an existing client" do
      RegistryMock
      |> expect(:lookup, fn _, _ -> [{1, 2}] end)

      GenServerMock
      |> expect(:call, fn 1, :key_establishment, 10_000 -> :result end)

      assert NTS.key_establishment("localhost", 1111) == :result
    end
  end
end
